# -*- coding: utf-8 -
#
# Copyright (c) 2008, 2009 Benoit Chesneau <benoitc@e-engura.com> 
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

"""
curl transport
"""

import re
import StringIO
import base64
import sys
from datetime import datetime, timedelta

import restclient
from restclient.errors import TransportError
from restclient.transport.base import *
from restclient.utils import to_bytestring, iri2uri

try:
    import pycurl
except ImportError:
    pycurl = None

import urllib
    
NORMALIZE_SPACE = re.compile(r'(?:\r\n)?[ \t]+')
def _normalize_headers(headers):
    return dict([ (key.lower(), NORMALIZE_SPACE.sub(value, ' ').strip())  for (key, value) in headers.iteritems()])


def _get_pycurl_errcode(symbol, default):
    """
    Returns the numerical error code for a symbol defined by pycurl.

    Different pycurl implementations define different symbols for error
    codes. Old versions never define some symbols (wether they can return the
    corresponding error code or not). The following addresses the problem by
    defining the symbols we care about.  Note: this allows to define symbols
    for errors that older versions will never return, which is fine.
    """
    return pycurl.__dict__.get(symbol, default)

if pycurl is not None:
    CURLE_COULDNT_CONNECT = _get_pycurl_errcode('E_COULDNT_CONNECT', 7)
    CURLE_COULDNT_RESOLVE_HOST = _get_pycurl_errcode('E_COULDNT_RESOLVE_HOST', 6)
    CURLE_COULDNT_RESOLVE_PROXY = _get_pycurl_errcode('E_COULDNT_RESOLVE_PROXY', 5)
    CURLE_GOT_NOTHING = _get_pycurl_errcode('E_GOT_NOTHING', 52)
    CURLE_PARTIAL_FILE = _get_pycurl_errcode('E_PARTIAL_FILE', 18)
    CURLE_SEND_ERROR = _get_pycurl_errcode('E_SEND_ERROR', 55)
    CURLE_WRITE_ERROR = _get_pycurl_errcode('E_WRITE_ERROR', 23)
    CURLE_SSL_CACERT = _get_pycurl_errcode('E_SSL_CACERT', 60)
    CURLE_SSL_CACERT_BADFILE = _get_pycurl_errcode('E_SSL_CACERT_BADFILE', 77)
    CURLE_ABORTED_BY_CALLBACK = _get_pycurl_errcode('E_ABORTED_BY_CALLBACK', 42)

class HTTPHeaderStream(StringIO.StringIO):
    def __init__(self, data_stream):
        StringIO.StringIO.__init__(self)
        self.data_stream = data_stream
        self.code = 0
        self.status = ''

    def write(self, string):
        StringIO.StringIO.write(self, string)

        # Check if header is HTTP status string
        match = re.search('^HTTP\/\d\.\d (\d+) (.+)', string)
        if match:
            self.code = int(match.groups()[0])
            self.status = match.groups()[1]
            if self.code >= 400:
                # Server returned error, close corresponding data stream
                self.data_stream.close()

class CurlTransport(HTTPTransportBase):
    """
    An HTTP transportthat uses pycurl.

    Pycurl is recommanded when you want fast access to http resources.
    We have added some basic management of authentification and proxies,
    but in case you want something specific you should use urllib2 or 
    httplib2 http clients. Any patch is welcome though ;)


    Here is an example to use authentification with curl httpclient :

    .. code-block:: python

        httpclient = CurlTransport()
        httpclient.add_credentials("test", "test")        

    .. seealso::

        `Pycurl <http://pycurl.sourceforge.net>`_
    """

    def __init__(self, timeout=None, proxy_infos=None):
        """ Curl transport constructor

        :param timeout: int, timeout of request
        :param proxy_infos: dict, infos to connect via proxy:

        .. code-block:: python

            {
                'proxy_user': 'XXXXXXX',
                'proxy_password': 'XXXXXXX',
                'proxy_host': 'proxy',
                'proxy_port': 8080,
            }
            
        """
        HTTPTransportBase.__init__(self, proxy_infos=proxy_infos)

        # path to certificate file
        self.cabundle = None

        if pycurl is None:
            raise RuntimeError('Cannot find pycurl library')

        self.timeout = timeout
        self.cookies = []
        self.csrf_token = None

    def _parseHeaders(self, header_file):
        header_file.seek(0)

        # Remove the status line from the beginning of the input
        unused_http_status_line = header_file.readline()
        lines = [line.strip() for line in header_file]

        # and the blank line from the end
        empty_line = lines.pop()
        if empty_line:
            raise TransportError("No blank line at end")

        self.cookies = []
        headers = {}
        for line in lines:
            if ":" in line:
                try:
                    name, value = line.split(':', 1)
                except ValueError:
                    raise TransportError(
                        "Malformed HTTP header line in response: %r" % (line,))

                value = value.strip()

                # HTTP headers are case-insensitive
                name = name.lower()
                headers[name] = value

                # preserve cookies
                if name == 'set-cookie':
                    match = re.match('^rest_access_token=(.+?);', value)
                    if match:
                        if self.csrf_token is None:
                            self.csrf_token = urllib.unquote(match.groups()[0])
                        else:
                            raise TransportError('User session has been reset')
                    else:
                        self.cookies.append(line)

        return headers

    def progress(self, download_t, download_d, upload_t, upload_d):
        # den 2012.02.06 Use progress callbacks to enable us to catch signals from hosting C code.
        #                just need to get into the python context from curl

        # dynamically update the timeout to support timeouts but also long file uploads
        if self.timeout is None:
            return 0

        if self.transfer_last_data_transfer_time is 0:
            self.transfer_last_data_transfer_time = datetime.now()

        # Ignore timeout for uploads
        if ((download_d > 0) and (self.transfer_last_download_d != download_d)) or \
            (upload_d > 0):
            self.transfer_last_download_d = download_d
            self.transfer_last_upload_d = upload_d
            self.transfer_last_data_transfer_time = datetime.now()
        else:
            delta = datetime.now() - self.transfer_last_data_transfer_time
            if delta.seconds > self.timeout:
                return 1        # timeout

        return 0

    # overrides base class method
    def clear_credentials(self):
        super(CurlTransport, self).clear_credentials()   # call the base class method
        # clear aE session token and any cross-site forgery token
        self.cookies = []
        self.csrf_token = None  
        
        
    def request(self, url, method='GET', body=None, headers=None, stream=None, user_agent=None):
        body = body or ""
        headers = headers or {}
        if user_agent is not None:
            headers['User-Agent'] = user_agent  # always set this value even if a previous value already exists

        # by default turn off default pragma
        headers.setdefault('Cache-control', 'max-age=0')
        headers.setdefault('Pragma', 'no-cache')

        # den 2012.01.15 Don't do 100-continue because Ruby f-s it up
        if method == 'PUT':
        #    headers.setdefault('Expect', '100-continue')
             headers['Expect'] = ''   # always set this value even if a previous value already exists

        # encode url
        origin_url = to_bytestring(url)
        url = iri2uri(origin_url)

        c = pycurl.Curl()
        try:
            # den 2012.02.06 improved timeout handling for large files
            c.setopt(pycurl.TIMEOUT, 0)
            self.transfer_last_data_transfer_time = 0
            self.transfer_last_download_d = 0
            self.transfer_last_upload_d = 0

            data = StringIO.StringIO()
            if stream is None:
                c.setopt(pycurl.WRITEFUNCTION, data.write)
                header = StringIO.StringIO()
            else:
                c.setopt(pycurl.WRITEFUNCTION, stream.write)
                header = HTTPHeaderStream(stream)

            c.setopt(pycurl.HEADERFUNCTION, header.write)
            c.setopt(pycurl.URL, url)
            c.setopt(pycurl.FOLLOWLOCATION, 1)
            c.setopt(pycurl.MAXREDIRS, 5)
            c.setopt(pycurl.NOSIGNAL, 1)
            if restclient.debuglevel > 0:
                c.setopt(pycurl.VERBOSE, 1)

            # den 2012.02.06 Progress callback used to catch sigints from hosting app
            c.setopt(c.NOPROGRESS, 0)
            c.setopt(c.PROGRESSFUNCTION, self.progress)

            # den 2012.01.15 Support self-signed certificate verification
            if self._get_allow_selfsigned():
                c.setopt(pycurl.SSL_VERIFYPEER, False)

            # automatic decompression
            c.setopt(pycurl.ENCODING, 'gzip,deflate')

            if self.cabundle:
                c.setopt(pycurl.CAINFO, self.cabundle)

            #set proxy
            if self.proxy_infos and self.proxy_infos.get('proxy_host', ''):
                c.setopt(pycurl.PROXYAUTH, pycurl.HTTPAUTH_ANY)
                c.setopt(pycurl.PROXY, self.proxy_infos.get('proxy_host'))

                proxy_port = self.proxy_infos.get('proxy_port', '')
                if proxy_port:
                    c.setopt(pycurl.PROXYPORT, str(proxy_port))

                user = self.proxy_infos.get('proxy_user', '')
                if user:
                    userpass = "%s:%s" % (user, self.proxy_infos.get('proxy_password', ''))
                    c.setopt(pycurl.PROXYUSERPWD, userpass)

            # authentification
            auth = self._get_credentials()
            user = auth.get('user', None)
            password = auth.get('password', None)
            if user is not None:
                c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_ANY)
                userpass = user + ':'
                if password is not None: # '' is a valid password
                    userpass += password
                
                # den 2012.01.15 Force inclusion of the credentials - we're HTTPS so it's ok to be basic
                # This is necessary because PUT / POST doesn't handle 401 authentication properly with
                # the way our ruby server works
                base = base64.encodestring(userpass.encode('utf-8')).strip()
                headers['authorization'] = "Basic %s" % base  # always set this value even if a previous value already exists: the credentials may have changed

                # den 2012.01.15 On pycurl Windows, for some reason this call generates a type exception
                # which doesn't make sense, but we're going to be forcing the inclusion of basic auth
                # so we don't need to make this call
                #c.setopt(pycurl.USERPWD, userpass)

            # erase previously set cookies
            c.setopt(pycurl.COOKIELIST, 'ALL')

            # set cookies returned by previous request
            for cookie in self.cookies:
                c.setopt(pycurl.COOKIELIST, cookie)

            # set CSRF token, if any
            if method != 'GET' and self.csrf_token:
                headers['X-CSRF-Token'] = self.csrf_token  # always set this value even if a previous value already exists

            # set method
            if method == "GET":
                c.setopt(pycurl.HTTPGET, 1)
            elif method == "HEAD":
                c.setopt(pycurl.HTTPGET, 1)
                c.setopt(pycurl.NOBODY, 1)
            elif method == "POST":
                c.setopt(pycurl.POST, 1)
            elif method == "PUT":
                c.setopt(pycurl.UPLOAD, 1)
            else:
                c.setopt(pycurl.CUSTOMREQUEST, method)

            if method in ('POST','PUT'):
                if hasattr(body, 'read'):
                    content_length = int(headers.pop('Content-Length', 0))
                    content = body
                else:
                    body = to_bytestring(body)
                    content = StringIO.StringIO(body)
                    if 'Content-Length' in headers:
                        del headers['Content-Length']
                    content_length = len(body)

                # den 2012.02.10 Add AEContent-Length in to allow detection of partially received files
                #       regular content-length >2GB can be stripped to blank by some intermediaries
                headers['AEContent-Length'] = "%d" % content_length

                if method == 'POST':
                    #c.setopt(pycurl.POSTFIELDSIZE, content_length)
                    c.setopt(pycurl.POSTFIELDSIZE_LARGE, content_length)
                else:
                    # den 2012.01.15 Support streaming >4GB files
                    #c.setopt(pycurl.INFILESIZE, content_length)
                    c.setopt(pycurl.INFILESIZE_LARGE, content_length)
                c.setopt(pycurl.READFUNCTION, content.read)

            if headers:
                _normalize_headers(headers)
                c.setopt(pycurl.HTTPHEADER,
                        ["%s: %s" % pair for pair in sorted(headers.iteritems())])

            try:
                c.perform()
            except pycurl.error, e:
                if e[0] not in (CURLE_SEND_ERROR, CURLE_WRITE_ERROR):
                    if restclient.debuglevel > 0:
                        print >>sys.stderr, str(e)
                    if e[0] == CURLE_ABORTED_BY_CALLBACK:
                        raise TransportError('Network operation timed out')
                    raise TransportError(e)

            response_headers = self._parseHeaders(header)
            code = c.getinfo(pycurl.RESPONSE_CODE)

            if code >= 400 and stream is not None:
                # Retrieve error description from header stream
                data.write('%d %s\n' % (header.code, header.status))

            return self._make_response(final_url=url, origin_url=origin_url,
                    status=code, headers=response_headers, body=data.getvalue())
        finally:
            c.close()

    def _make_response(self, final_url=None, origin_url=None, status=None, 
            headers=None, body=None):
        infos = headers or {}    
        final_url = infos.get('location', final_url)
        infos.update({
            'status': status,
            'final_url': final_url,
            'origin_url': origin_url
        })
        resp = HTTPResponse(infos)
        return resp, body

    def curlVersion(self):
        if pycurl is None:
            return 'Cannot find pycurl library'
        else:
            return pycurl.version

